import { DSVModel } from 'tde-csvviewer';
import { MutableDataModel, DataModel, SelectionModel } from 'tde-datagrid';
import { Litestore } from './litestore';
import { DSVEditor } from './widget';
import { Signal } from '@lumino/signaling';
import { ListField } from 'tde-datastore';
export declare class EditorModel extends MutableDataModel {
    private _clipboard;
    private _litestore;
    private _model;
    private _rowsAdded;
    private _columnsAdded;
    private _onChangeSignal;
    private _rowsRemoved;
    private _columnsRemoved;
    private _saving;
    constructor(options: DSVModel.IOptions);
    /**
     * The model which holds the string containing the contents of the file.
     */
    get model(): DSVModel;
    /**
     * A signal that emits when data is set to notify the Editor.
     */
    get onChangedSignal(): Signal<this, DSVEditor.ModelChangedArgs>;
    /**
     * The datastore used by the this class to read properties.
     */
    get litestore(): Litestore;
    /**
     * The setter used by parent classes to intialize the litestore.
     */
    set litestore(value: Litestore);
    /**
     * The grid's current number of rows by region.
     *
     */
    rowCount(region: DataModel.RowRegion): number;
    /**
     * The grid's current number of columns by region.
     *
     * Note: the UI components use this method to get the column count
     * so it should reflect the grid's columns.
     */
    columnCount(region: DataModel.ColumnRegion): number;
    /**
     * The grid's current number of rows. TOTAL, NOT BY REGION.
     */
    totalRows(): number;
    /**
     * The grid's current number of columns. TOTAL, NOT BY REGION.
     *
     * Notes: This is equivalent to columnCount('body')
     */
    totalColumns(): number;
    /**
     * This function is called by the datagrid to fill in values. It is called many times
     * and so should be efficient.
     */
    data(region: DataModel.CellRegion, row: number, column: number): any;
    /**
     * The method for setting data in a single cell.
     */
    setData(region: DataModel.CellRegion, row: number, column: number, values: any, rowSpan?: number, columnSpan?: number, update?: DSVEditor.ModelChangedArgs | null): boolean;
    /**
     * The method for setting data in a range of cells.
     */
    bulkSetData(rowArray: Array<number>, columnArray: Array<number>, value: string, startRow: number, startColumn: number, endRow: number, endColumn: number): void;
    /**
     * Add rows to the Editor.
     */
    addRows(region: DataModel.CellRegion, start: number, span?: number): DSVEditor.ModelChangedArgs;
    /**
     * Add columns to the editor.
     */
    addColumns(region: DataModel.CellRegion, start: number, span?: number): DSVEditor.ModelChangedArgs;
    /**
     * Add rows to the editor.
     */
    removeRows(region: DataModel.CellRegion, start: number, span?: number): DSVEditor.ModelChangedArgs;
    /**
     * Remove rows from the editor.
     */
    removeColumns(region: DataModel.CellRegion, start: number, span?: number): DSVEditor.ModelChangedArgs;
    /**
     * Move rows in the grid.
     */
    moveRows(region: DataModel.CellRegion, start: number, end: number, span: number, update: DSVEditor.ModelChangedArgs): DSVEditor.ModelChangedArgs;
    /**
     * Handles the computations involved in moving rows in the grid.
     */
    rowMapSplice(rowMap: ListField.Value<number>, start: number, end: number, span: number): ListField.Splice<number>[];
    /**
     * Move columns in the grid.
     */
    moveColumns(region: DataModel.CellRegion, start: number, end: number, span: number, update: DSVEditor.ModelChangedArgs): DSVEditor.ModelChangedArgs;
    /**
     * Handles computations associated with moving columns in the grid.
     */
    columnMapSplice(columnMap: ListField.Value<number>, start: number, end: number, span: number): ListField.Splice<number>[];
    /**
     * Clear a region of cells in the grid.
     */
    clearCells(region: DataModel.CellRegion, selection: SelectionModel.Selection): DSVEditor.ModelChangedArgs;
    /**
     * Clear rows in the grid.
     */
    clearRows(region: DataModel.CellRegion, start: number, span?: number): DSVEditor.ModelChangedArgs;
    /**
     * Clear columns in the grid.
     */
    clearColumns(region: DataModel.CellRegion, start: number, span?: number): DSVEditor.ModelChangedArgs;
    /**
     * Cut a selection of cells.
     * NOTE: this method both copies the cells to the _clipboard property and clears them
     * from the region.
     *
     * TODO: refactor so that cut uses the already existing method clearCells.
     */
    cut(region: DataModel.CellRegion, startRow: number, startColumn: number, endRow: number, endColumn: number): DSVEditor.ModelChangedArgs;
    /**
     * Copies a selection of data to the local propery _clipboard.
     * NOTE: this does not copy to the system clipboard. For this use the DataGrid method
     * copyToClipboard
     */
    copy(region: DataModel.CellRegion, startRow: number, startColumn: number, endRow: number, endColumn: number): void;
    /**
     * Paste a selection of cells onto the grid.
     * NOTE: this method first checks if we have data stored in our local clipboard. It
     * only checks for the data parameter if the local clipboard is empty.
     */
    paste(region: DataModel.CellRegion, row: number, column: number, data?: string | null): DSVEditor.ModelChangedArgs;
    /**
     * Emits the change which undoes the change passed in.
     */
    emitOppositeChange(change: DataModel.ChangedArgs): void;
    /**
     * Emits the change which is passed in as an argument.
     */
    emitCurrentChange(change: DataModel.ChangedArgs): void;
    /**
     * Returns the serializes string with the changes added.
     */
    updateString(): string;
    /**
     * Produces two arrays inverseRowMap and inverseColumnMap such that for each x
     * in rowMap, inverseRowMap[x] === rowMap.indexOf(x). Likewise for inverseColumnMap
     * and columnMap.
     */
    private _invertMaps;
    /**
     * translate from the Grid's row IDs to our own standard
     */
    private _absoluteIndex;
    /**
     * translate from our unique row ID to the Grid's standard
     */
    private _regionIndex;
    /**
     * Processes updates from the DSVModel.
     */
    private _receiveModelSignal;
    /**
     * Adds new rows coming in from the asynchronous parsing of string by the DSVModel.
     */
    private _assimilateNewRows;
    /**
     * The total rows currently stored in the DSVModel.
     */
    private _modelRows;
    /**
     * The total columns currently stored in the DSVModel.
     */
    private _modelColumns;
    /**
     * Computes the offset index at the end of a given row (note: row delimeter not included).
     */
    private _rowEnd;
    private _openSlice;
    private _columnSlicePattern;
    private _performMacroSlice;
    private _serializer;
    /**
     * Returns a blank row with the correct numbers of columns and correct delimiters
     * @param model The DSV model being used
     * @param row The index of the row being inserted (determines whether to add a row delimiter or not)
     */
    private _blankRow;
    private _peformMicroSlice;
}
export declare type MapUpdate = 'add' | 'remove' | 'move' | 'clear';
export declare type SlicePattern = {
    buffers: Array<string>;
    slices: Array<Array<number>>;
};
